# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of PrefixCache with LRU cache and Trie based key lookup.

i) Initialize cache

# Prefix return by prefill function of mixtral-8x22b model with max_prefill_length=1024,
# int8 quantize_kvcache is 235_930_060 bytes, nearly 256 MB.
# It need about 256 * 256 MB = 64GB to caching 256 prompts prefill.
hbm_bytes = 64_000_000_000  # 64 GB
dram_bytes = 640_000_000_000  # 640 GB
prefix_cache = PrefixCache(hbm_bytes=hbm_bytes, dram_bytes=dram_bytes)

ii) Read from Cache:

# From request
prompt: str = "blah"
full_key: Key = tokenizer.tokenize(prompt)

# Inside prefill call
orig_key_len = len(full_key)

matched_key = prefix_cache.fetch_longest_common_prefix_key(key)
if matched_key is not None and load_cache_is_efficient_enough(matched_key):
  cache_results = prefix_cache.load(matched_key)
else:
  cache_results = None

if cache_results:
   # load cache successfully
   matched_len = calculate_matched_len(full_key, matched_key)
   cached_prefix = cache_results.prefix
else:
   cached_prefix = None

iii) Run prefill un-cached prompt and write to cache

full_kv_cache = prefill(prompt, cached_prefix, matched_len)

# if the cache didn't contain prefix or if it was a partial match
if cached_prefix is None or matched_len != orig_key_len:
    prefix_cache.save(full_key, Value(
        prefix=full_kv_cache,
        true_length=true_length,
        padded_length=padded_length,
        tokens=full_key,
    ))

"""

from collections import OrderedDict
from typing import Any, Optional, Tuple
import abc
import dataclasses
import jax
import jax.numpy as jnp
import logging
import threading

logger = logging.getLogger(__name__)

Token = int
# Tuple of tokens from prompt
Key = Tuple[Token, ...]
Prefix = Any  # KVCache for one prompt


class Value:
  """Object stored contains the actual KVcache

  Attributes:
    prefix:
      Readonly. Prefix Cache using in model. Should be pytree of all jax.Array.
    true_length:
      Readonly. True length of tokens calculate prefix. Should be <= than len(tokens).
      true_length will be min(true_length, len(tokens))
    padded_length:
      Readonly. Length of tokens including padding calculate prefix.
    tokens:
      Readonly. Tokens calculate prefix. may include partial of padding.
    prefix_size_bytes:
      Readonly. bytes of prefix.
    device:
      Readonly. Devices of prefix. The same structure of pytree to prefix.
      The device may be different from actually prefix is in.
      It used for retrieved back to original device.
  """

  def __init__(
      self,
      *,
      prefix: Prefix,
      true_length: int,
      padded_length: int,
      tokens: tuple[Token, ...],
      prefix_size_bytes: Optional[int] = None,
      device=None,
  ):
    """Init Value with attributes.

    If true_length shorter than len(tokens), true_length will adjust to len(tokens).
    If prefix_size_bytes is not provided, calculate automatically.
    prefix should be pytree of all jax.Array. It may raise exception if there is anything not a jax.Array,
    If either prefix_size_bytes and device is None, get from prefix.
    """
    self._prefix = prefix
    self._true_length = self._maybe_adjust_true_length(true_length, tokens)
    self._padded_length = padded_length
    self._tokens = tokens
    if prefix_size_bytes is None:
      self._prefix_size_bytes: int = jax.tree.reduce(
          lambda acc, array: acc + array.nbytes,
          prefix,
          0,
      )
    else:
      self._prefix_size_bytes = prefix_size_bytes

    if device is None:
      self._device = jax.tree.map(lambda x: x.device, prefix)
    else:
      self._device = device

  @property
  def prefix(self) -> Prefix:
    return self._prefix

  @property
  def true_length(self) -> int:
    return self._true_length

  @property
  def padded_length(self) -> int:
    return self._padded_length

  @property
  def tokens(self) -> tuple[Token, ...]:
    return self._tokens

  @property
  def prefix_size_bytes(self) -> int:
    return self._prefix_size_bytes

  @property
  def device(self) -> int:
    return self._device

  def __eq__(self, other: Any) -> bool:
    if not isinstance(other, Value):
      return False
    return (
        other.padded_length == self.padded_length
        and other.tokens == self.tokens
        and jax.tree.all(jax.tree.map(jnp.array_equal, other.prefix, self.prefix))
        and other.prefix_size_bytes == self.prefix_size_bytes
    )

  def _maybe_adjust_true_length(self, true_length: int, tokens: tuple[Token, ...]) -> int:
    if true_length > len(tokens):
      logger.warning("true_length=%d should <= len(tokens)=%d.", true_length, len(tokens))

    return min(true_length, len(tokens))


def device_put_value(value: Value, device: Any = None) -> Value:
  """Create a new value with prefix put to device.

  If the device is the same as value.prefix, we expect no copy here in jax.device_put.

  Args:
    value: Value to put.
    device:
      The same as the jax.device_put device to put the value.prefix.
      if None, put to the value.device.
  Returns:
    Values with prefix put to device.
  """
  put_device = device
  if put_device is None:
    put_device = value.device
  return Value(
      prefix=jax.device_put(value.prefix, put_device),
      true_length=value.true_length,
      padded_length=value.padded_length,
      tokens=value.tokens,
      prefix_size_bytes=value.prefix_size_bytes,
      device=value.device,
  )


class PrefixCacheTrie:
  """Stores prefix tokens as a trie for fast lookup index of PrefixCache store in cache.

  Insert longer Key replace shorter key to be the longest common prefix key.
  The shorter key will never be returned even if longer key is erased, and should got evicted in the future.

  Assume Key is equal length to tokens, which can be used to slice prompt and cache Value.
  Should check the return key common prefix length by the caller.

  If erase the Key not the leaf, nothing will happen.
  If erased key match at a leaf, delete the node and ancestors would be the leaf after deleted.
  """

  @dataclasses.dataclass
  class Node:
    """Trie Node."""

    parent: Optional["PrefixCacheTrie.Node"] = None
    token: Optional[Token] = None
    children: dict[Token, "PrefixCacheTrie.Node"] = dataclasses.field(default_factory=dict)

    def is_leaf(self):
      return len(self.children) == 0

    def get_one_child_token(self) -> Optional[Token]:
      if len(self.children) == 0:
        return None
      return next(iter(self.children.keys()))

  def __init__(self):
    self._saved_keys: list[Key] = []
    self._root = PrefixCacheTrie.Node()

  def insert(self, key: Key):
    """Insert key into the trie."""
    node = self._root
    for token in key:
      if token not in node.children:
        node.children[token] = PrefixCacheTrie.Node(parent=node, token=token)
      node = node.children[token]

  def get_longest_common_prefix_key(self, key: Key) -> Optional[Key]:
    """Get the key with longest common prefix.
    If not found at least one token match, return None."""
    result_tokens: list[Token] = []

    node = self._root
    for token in key:
      if token not in node.children:
        break
      node = node.children[token]
      result_tokens.append(token)

    if len(result_tokens) == 0:
      return None

    while not node.is_leaf():
      token = node.get_one_child_token()
      if token is None:
        break
      result_tokens.append(token)
      node = node.children[token]

    return tuple(result_tokens)

  def erase(self, key: Key) -> None:
    """Erase key in trie if it is leaf."""
    node = self._root
    for token in key:
      if token not in node.children:
        return
      node = node.children[token]

    while node.is_leaf():
      parent = node.parent
      if parent is None or node.token not in parent.children:
        return
      del parent.children[node.token]
      node = parent


class ValueStorageInterface(abc.ABC):
  """Interface for Value storage."""

  @abc.abstractmethod
  def get_max_size_bytes(self) -> int:
    """Get the max size bytes in storage."""

  @abc.abstractmethod
  def has_enough_space(self, needed_bytes: int) -> bool:
    """Calculate if needed_bytes size can add to storage."""

  @abc.abstractmethod
  def add(self, key: Key, value: Value) -> bool:
    """Add value and return True. If storage is full, return False."""

  @abc.abstractmethod
  def retrieve(self, key: Key, device: Any = None) -> Optional[Value]:
    """Return value from storage or None if not found.

    Args:
      key: key to retrieve value.
      device:
        The same as device in jax.device_put. Retrieve the value to device.
        If device is None, retrieve the value to it's original device while saved.
    Returns:
      Value retrieved from storage or None if not found.
    """

  @abc.abstractmethod
  def evict(self, key: Key) -> Optional[Value]:
    """Evict and return value, or None if key is not in storage."""

  @abc.abstractmethod
  def contains(self, key: Key) -> bool:
    """If there is key in storage."""


class BasicStorage:
  """Basic implement calculating size and save value into dict without modify."""

  def __init__(self, max_size_bytes: int):
    """
    Args:
      max_size_bytes: Maximum bytes use
    """
    self._max_size_bytes = max_size_bytes
    self._remain_size_bytes = max_size_bytes
    self._saved_values: dict[Key, Value] = {}

  def get_max_size_bytes(self) -> int:
    return self._max_size_bytes

  def has_enough_space(self, needed_bytes: int) -> bool:
    """Calculate if needed_bytes size can add to storage."""
    return self._remain_size_bytes >= needed_bytes

  def add(self, key: Key, value: Value) -> bool:
    """Add value and return True. If storage is full, return False.

    The value will not copied. Be aware not to modify the value after add to storage.
    Storage is expected to have enough space.
    """
    if not self.has_enough_space(value.prefix_size_bytes):
      logger.warning(
          "should check enough space before add to storage, but remain=%d not enough for value=%d",
          self._remain_size_bytes,
          value.prefix_size_bytes,
      )
      return False

    self._saved_values[key] = value
    self._remain_size_bytes -= value.prefix_size_bytes
    return True

  def retrieve(self, key: Key) -> Optional[Value]:
    """Return value from storage or None if not found.

    Be aware the storage is not return a copy. Clone the Value first if additional modification needed.
    """
    if key not in self._saved_values:
      logger.warning("key=%r should exist in storage before retrieve, but not found", key)
      return None
    return self._saved_values[key]

  def evict(self, key: Key) -> Optional[Value]:
    """Evict and return value, or None if key is not in storage.
    Key is expected to be found.
    """
    if key not in self._saved_values:
      logger.warning("key=%r should exist in storage before evict, but not found", key)
      return None
    value = self._saved_values.pop(key)
    self._remain_size_bytes += value.prefix_size_bytes
    return value

  def contains(self, key: Key) -> bool:
    """If there is key in storage."""
    return key in self._saved_values


class HBMStorage(ValueStorageInterface):
  """Stores kv storage values in HBM.

  Store the Value into the specific HBM device, which is the same type as device in jax.device_put.
  The Value would be jax.device_put to the HBM device after add, and retrieve back to the original device.
  """

  def __init__(self, max_size_bytes: int, device: Any = None):
    """Init the HBMStorage with max size limit and device to store the Value.

    Args:
      max_size_bytes: Maximum bytes of HBM to use for storage
      device:
        the same type as jax.device_put. It is used to store the cache Value.
        If None, do not move the Value.
    """
    self._storage = BasicStorage(max_size_bytes)
    self._device = device

  def get_max_size_bytes(self) -> int:
    return self._storage.get_max_size_bytes()

  def has_enough_space(self, needed_bytes: int) -> bool:
    """Calculate if needed_bytes size can add to storage."""
    return self._storage.has_enough_space(needed_bytes)

  def add(self, key: Key, value: Value) -> bool:
    """Add key/value pair into the cache.

    Depend on jax.device_put,
    the Value will not be copied if the device storing the cache is the same as the origin device of Value.
    Storage is expected to have enough space.

    Args:
      key: key of cache index.
      value: Value to store.
    Returns:
      True if successful. False if failed due to not enough space.
    """
    hbm_value = device_put_value(value, self._device)
    return self._storage.add(key, hbm_value)

  def retrieve(self, key: Key, device: Any = None) -> Optional[Value]:
    """Retrieve value back to the original device or None if not found.

    Be aware the storage may not return a copy if the original devices is the same as depend on jax.device_put.
    Key is expected to be found.
    """
    hbm_value = self._storage.retrieve(key)
    if hbm_value is None:
      return None

    return device_put_value(hbm_value, device)

  def evict(self, key: Key) -> Optional[Value]:
    """Evict and return value, or None if key is not in storage.
    Key is expected to be found.
    """
    return self._storage.evict(key)

  def contains(self, key: Key) -> bool:
    """If there is key in storage."""
    return self._storage.contains(key)


class DRAMStorage(ValueStorageInterface):
  """Stores KV Cache values in host DRAM."""

  def __init__(self, max_size_bytes: int):
    """
    Args:
      max_size_bytes: Maximum bytes of host DRAM to use for storage
    """
    self._storage = BasicStorage(max_size_bytes)

  def get_max_size_bytes(self) -> int:
    return self._storage.get_max_size_bytes()

  def has_enough_space(self, needed_bytes: int) -> bool:
    """Calculate if needed_bytes size can add to storage."""
    return self._storage.has_enough_space(needed_bytes)

  def add(self, key: Key, value: Value) -> bool:
    """Add value into host DRAM.

    Return false if storage does not have enough space.
    Do not use this function to check if has enough space.
    This function will first move to host DRAM before check the space.
    The storage will copy to the host DRAM if originally on device,
    or with the same reference to the value if originally on host.
    Do not use the value after this function if originally on host since the value will not copy.
    """
    host_value = Value(
        prefix=jax.device_get(value.prefix),
        true_length=value.true_length,
        padded_length=value.padded_length,
        tokens=value.tokens,
        prefix_size_bytes=value.prefix_size_bytes,
        device=value.device,
    )

    return self._storage.add(key, host_value)

  def retrieve(self, key: Key, device: Any = None) -> Optional[Value]:
    """Return value from storage to the original device or None if not found.

    If the original device save in the storage is cpu, the storage will not copied.
    Do not modify the storage prefix retrieved.
    """
    host_value = self._storage.retrieve(key)
    if host_value is None:
      return None

    return device_put_value(host_value, device)

  def evict(self, key: Key) -> Optional[Value]:
    """Evict and return value, or None if key is not in storage."""
    return self._storage.evict(key)

  def contains(self, key: Key) -> bool:
    """If there is key in storage."""
    return self._storage.contains(key)


class LRUStrategy:
  """Least recently used cache strategy manage key."""

  def __init__(self):
    self._order: OrderedDict[Key, None] = OrderedDict()

  def evict(self) -> Optional[Key]:
    """Return and pop the least recently used key."""
    if len(self._order) == 0:
      return None
    return self._order.popitem(last=False)[0]

  def use(self, key: Key) -> None:
    """Updated the usage history."""
    if key not in self._order:
      self._order[key] = None
    else:
      self._order.move_to_end(key, last=True)


@dataclasses.dataclass
class StorageWithStrategy:
  """Storage with corresponding strategy"""

  storage: ValueStorageInterface
  strategy: LRUStrategy


class HierarchicalCache:
  """Hierarchical Cache contains two layers of ValueStorageInterface.

  The first layer contains subset fo key / value pairs of the second layer.
  The second storage max size bytes should >= first storage max size bytes.
  Use LRU for each layer.
  Add the Value will save to all layers.
  Retrieve the Value will retrieve to HBM and then saved to all layers.
  The added value size should less than the first layer max size.
  If the first layer max size cannot contains the added Value, add will failed.
  """

  def __init__(self, layers: Tuple[ValueStorageInterface, ValueStorageInterface]):
    assert (
        layers[0].get_max_size_bytes() <= layers[1].get_max_size_bytes()
    ), "Bottom layer of storage need to be larger than top."

    self._layers = [StorageWithStrategy(storage, LRUStrategy()) for storage in layers]

  def add(self, key: Key, value: Value) -> tuple[bool, dict[Key, Value]]:
    """Add to all layers and return (ok, dict[fully evicted from hierarchical cache key value pair]).

    Beware in some error case, there may be not ok but have some evicted key value.
    """
    needed_bytes = value.prefix_size_bytes
    if self._layers[0].storage.get_max_size_bytes() < needed_bytes:
      logging.warning(
          "Trying to add value larger than top layer max size. need_bytes=%d, max_size_bytes=%d",
          needed_bytes,
          self._layers[0].storage.get_max_size_bytes(),
      )
      return False, {}

    # Only return last layers evicted key value pair which is fully evicted from hierarchical cache.
    all_ok = True
    last_layer_evicted_key_values: dict[Key, Value] = {}
    for layer in self._layers:
      if layer.storage.contains(key):
        last_layer_evicted_key_values = {}
        continue

      ok, last_layer_evicted_key_values = self._evict_to_enough_space(layer, needed_bytes)
      all_ok = all_ok and ok

    if not all_ok:
      logging.error("Cannot evict enough space after checking max_size is enough for bytes=%d.", needed_bytes)
      return False, last_layer_evicted_key_values

    for layer in self._layers:
      if not layer.storage.contains(key):
        if not layer.storage.add(key, value):
          logging.error("Cannot add to storage. key=%r, needed_bytes=%d", key, needed_bytes)
          return False, last_layer_evicted_key_values

      layer.strategy.use(key)

    return True, last_layer_evicted_key_values

  def retrieve(self, key: Key, device: Any = None) -> Optional[Value]:
    """Retrieve from all layers and add to all layers.

    Args:
      key: key to retrieve.
      device:
        The same type as the device in jax.device_put. Return the Value put on the device.
        If None, the Value will be put on the Value.device.
    Returns:
      Value retrieved from all layers or None if not found.
      The Value.device is not changed to device retrieved.
    """
    value: Optional[Value] = None
    for layer in self._layers:
      if layer.storage.contains(key):
        value = layer.storage.retrieve(key, device)
        break

    if value is None:
      logging.warning("Should check key exist before retrieve, but fail for key=%r", key)
      return None

    for layer in self._layers:
      if not layer.storage.contains(key):
        if not self._evict_to_enough_space(layer, value.prefix_size_bytes):
          logging.error("Cannot evict enough space for retrieved Value to other layers.")
          continue

        if not layer.storage.add(key, value):
          logging.error("Cannot add retrieved Value to other layers.")
          continue

      layer.strategy.use(key)

    return value

  def _evict_to_enough_space(self, layer: StorageWithStrategy, needed_bytes: int) -> tuple[bool, dict[Key, Value]]:
    """Evict layer to enough bytes for add and return (ok, dict[evicted key, evicted value])."""
    evicted_key_values: dict[Key, Value] = {}
    while not layer.storage.has_enough_space(needed_bytes):
      evicted_key = layer.strategy.evict()
      if evicted_key is None:
        logging.error("Cannot evict enough space for bytes=%d.", needed_bytes)
        return False, evicted_key_values

      evicted_value = layer.storage.evict(evicted_key)
      if evicted_value is None:
        logging.error("Key should in storage before evict but not. key=%r", evicted_key)
        continue

      evicted_key_values[evicted_key] = evicted_value

    return True, evicted_key_values


class PrefixCache:
  """Store Prefix KV cache.

  Use hierarchical cache of two layers the first in the HBM and the second in the host DRAM.
  Assuming HBM is available, or the cache would degrade to two layers on DRAM.
  If cache is full, evict least-recently used entries (LRU).
  LRU strategy is apply to all layers.
  The cache in HBM will be subset of the cache in host DRAM.
  For example:
    For HBM can contain 2 values, and DRAM can contain 5 values,
    [1, 2, 3, 4, 5] LRU history, the [1, 2] will in HBM, and [1, 2, 3, 4, 5] will in DRAM.
  Always return cache after load into HBM.
  The value need to be <= to the max size in HBM.
  DRAM max size need to be >= than HBM max size.
  """

  def __init__(self, hbm_bytes: int, dram_bytes: int):
    """
    dram_bytes >= hbm_bytes
    Args:
      hbm_bytes: Total amount of HBM to use for cache.
      dram_bytes: Total amount of DRAM to use for cache.
    """
    # TODO(yuyanpeng): way to disable DRAM cache
    assert dram_bytes >= hbm_bytes, "DRAM max size need to be >= than HBM max size."
    self._lock = threading.Lock()
    self._hbm_bytes = hbm_bytes
    self._dram_bytes = dram_bytes
    # init in clear()
    self._trie: PrefixCacheTrie
    self._cache: HierarchicalCache
    self.clear()

  def fetch_longest_common_prefix_key(self, key: Key) -> Optional[Key]:
    """Returns key with longest common prefix matched or None if not found."""
    logger.debug("fetch_longest_common_prefix_key, key=%r", key)
    with self._lock:
      matched_key = self._trie.get_longest_common_prefix_key(key)
      logger.debug("matched_key=%r", matched_key)
      return matched_key

  def save(self, key: Key, value: Value) -> bool:
    """Save key/value to the cache."""
    logger.debug("save key=%r", key)
    with self._lock:
      ok, evicted = self._cache.add(key, value)
      for evicted_key in evicted.keys():
        self._trie.erase(evicted_key)
      if not ok:
        logger.warning("Cannot add to cache")
        return False
      self._trie.insert(key)
      return True

  def load(self, key: Key, device: Any = None) -> Optional[Value]:
    """Returns Value stored with key or None if not found.

    Args:
      key: key to load.
      device:
        The same type as device in the jax.device_put. Load the Value on the device.
        If None, load the Value on the original device Value.device.
    Return:
      Value stored with key or None if not found.
      The Value.device is not changed to device loaded on.
    """
    logger.debug("load key=%r", key)
    with self._lock:
      value = self._cache.retrieve(key, device)
      if value is None:
        logger.warning(
            "The key should fetched by fetch_longest_common_prefix_key, load key=%r should be valid but not.", key
        )
        return None
      return value

  def clear(self):
    """Clear entire cache."""
    logger.debug("clear cache")
    with self._lock:
      self._trie = PrefixCacheTrie()
      self._cache = HierarchicalCache(
          layers=(
              HBMStorage(self._hbm_bytes),
              DRAMStorage(self._dram_bytes),
          )
      )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration test of the orchestrator.

This test tests the multithreaded orchestrator, where a prefill request is
popped onto a prefill queue, prefilled, sent to a generation queue and run for
a number of decoding steps.

In operation, it will use gRPC so we can 'yield' from the function to get return
values in the same way that they would be streamed back.

Similar to 'mock_engine_test' we can use known token values and a singleton
weight to test our operation.

Let the prefill engine have a weight of [2] and the generate engine have a
weight of [3].

I.e. if we prefill [2, 65, 66] (i.e. <BOS>, 'A', 'B') using an ACII vocab,
we should get [4, 130, 132].

If we then insert that and run three generation steps, we should see
266+0 / 2 = 266
266 + [266] / 4  = 332
266 + [266, 332] / 4 = 415
I.e. ['Ċ', 'Ō', 'Ɵ'] when converted back with chr().

Therefore we should get back the character sequence '$lǔ' if we request 3 tokens
decoded (these are the ascii chars at those indices which is what the test
tokenizer returns).
"""

import unittest
from parameterized import parameterized
from jetstream.core import orchestrator
from jetstream.core.proto import jetstream_pb2
from jetstream.core.utils.return_sample import ReturnSample
from jetstream.engine import mock_engine


class OrchestratorTest(unittest.IsolatedAsyncioTestCase):

  def _setup_driver(
      self, interleaved_mode: bool = True, multi_sampling: bool = False
  ):
    prefill_engine = mock_engine.TestEngine(
        batch_size=32, cache_length=256, weight=2.0
    )
    # Create a generate engine with a different set of weights
    # so that we can test that the right one is in use at a given time.
    generate_engine = mock_engine.TestEngine(
        batch_size=4, cache_length=32, weight=4.0
    )
    driver = orchestrator.Driver(
        prefill_engines=[prefill_engine],
        generate_engines=[generate_engine],
        prefill_params=[prefill_engine.load_params()],
        generate_params=[generate_engine.load_params()],
        interleaved_mode=interleaved_mode,
        multi_sampling=multi_sampling,
    )
    return driver

  def _setup_driver_chunked_prefill(self, interleaved_mode: bool = True):
    prefill_engine = mock_engine.TestEngine(
        batch_size=32, cache_length=256, weight=2.0, use_chunked_prefill=True
    )
    # Create a generate engine with a different set of weights
    # so that we can test that the right one is in use at a given time.
    generate_engine = mock_engine.TestEngine(
        batch_size=4,
        cache_length=32,
        weight=4.0,
        use_chunked_prefill=True,
    )
    driver = orchestrator.Driver(
        prefill_engines=[prefill_engine],
        generate_engines=[generate_engine],
        prefill_params=[prefill_engine.load_params_dict()],
        generate_params=[generate_engine.load_params()],
        interleaved_mode=interleaved_mode,
    )
    return driver

  @parameterized.expand([True, False])
  async def test_orchestrator_chunked_prefill(self, interleaved_mode: bool):
    """Test the multithreaded orchestration."""
    driver = self._setup_driver_chunked_prefill(interleaved_mode)
    client = orchestrator.LLMOrchestrator(driver=driver)

    # The string representation of np.array([[65, 66]]), [2] will be prepend
    # as BOS.
    text = "AB"

    request = jetstream_pb2.DecodeRequest(
        text_content=jetstream_pb2.DecodeRequest.TextContent(text=text),
        max_tokens=3,
    )
    iterator = client.Decode(request)
    # chr of [135, 168, 210].
    expected_text = ["\x87", "¨", "Ò", ""]
    expected_token_ids = [135, 168, 210, None]
    counter = 0
    async for resp in iterator:
      output_text = resp.stream_content.samples[0].text
      token_ids = resp.stream_content.samples[0].token_ids
      output_token_id = token_ids[0] if len(token_ids) > 0 else None
      print(f"actual output: {output_text=} {output_token_id=}")
      assert output_text == expected_text[counter]
      assert output_token_id == expected_token_ids[counter]
      counter += 1
    driver.stop()
    print("Orchestrator driver stopped.")

  @parameterized.expand([True, False])
  async def test_orchestrator(self, interleaved_mode: bool):
    """Test the multithreaded orchestration."""
    driver = self._setup_driver(interleaved_mode)
    client = orchestrator.LLMOrchestrator(driver=driver)

    # The string representation of np.array([[65, 66]]), [2] will be prepend
    # as BOS.
    text = "AB"

    request = jetstream_pb2.DecodeRequest(
        text_content=jetstream_pb2.DecodeRequest.TextContent(text=text),
        max_tokens=3,
    )
    iterator = client.Decode(request)
    # chr of [266, 332, 415].
    expected_text = ["Ċ", "Ō", "Ɵ", ""]
    expected_token_ids = [266, 332, 415, None]
    counter = 0
    async for resp in iterator:
      output_text = resp.stream_content.samples[0].text
      token_ids = resp.stream_content.samples[0].token_ids
      output_token_id = token_ids[0] if len(token_ids) > 0 else None
      print(f"actual output: {output_text=} {output_token_id=}")
      assert output_text == expected_text[counter]
      assert output_token_id == expected_token_ids[counter]
      counter += 1
    driver.stop()
    print("Orchestrator driver stopped.")

  @parameterized.expand([1, 2, 3, 4])
  async def test_orchestrator_multi_sampling(self, num_samples: int):
    """Test the multithreaded orchestration."""
    driver = self._setup_driver(interleaved_mode=True, multi_sampling=True)
    client = orchestrator.LLMOrchestrator(driver=driver)

    # The string representation of np.array([[65, 66]]), [2] will be prepend
    # as BOS.
    text = "AB"

    request = jetstream_pb2.DecodeRequest(
        text_content=jetstream_pb2.DecodeRequest.TextContent(text=text),
        max_tokens=3,
        num_samples=num_samples,
    )
    iterator = client.Decode(request)
    # chr of [266, 332, 415].
    expected_text = ["Ċ", "Ō", "Ɵ", ""]
    expected_token_ids = [266, 332, 415, None]
    counter = 0
    async for resp in iterator:
      for sample in resp.stream_content.samples:
        output_text = sample.text
        token_ids = sample.token_ids
        output_token_id = token_ids[0] if len(token_ids) > 0 else None
        print(f"actual output: {output_text=} {output_token_id=}")
        assert output_text == expected_text[counter]
        assert output_token_id == expected_token_ids[counter]
      counter += 1
    driver.stop()
    print("Orchestrator driver stopped.")

  @parameterized.expand([True, False])
  async def test_orchestrator_client_tokenization_chunked_prefill(
      self, interleaved_mode: bool
  ):
    """Test the multithreaded orchestration."""
    driver = self._setup_driver_chunked_prefill(interleaved_mode)
    client = orchestrator.LLMOrchestrator(driver=driver)

    # The token ids of  string "AB", [2] will be prepend
    # as BOS.
    token_ids = [65, 66]

    request = jetstream_pb2.DecodeRequest(
        token_content=jetstream_pb2.DecodeRequest.TokenContent(
            token_ids=token_ids
        ),
        max_tokens=3,
    )
    iterator = client.Decode(request)
    # Return token ids only when in client side tokenization mode.
    expected_text = ["", "", "", ""]
    expected_token_ids = [133, 166, 207, None]
    counter = 0
    async for resp in iterator:
      output_text = resp.stream_content.samples[0].text
      token_ids = resp.stream_content.samples[0].token_ids
      output_token_id = token_ids[0] if len(token_ids) > 0 else None
      print(f"actual output: {output_text=} {output_token_id=}")
      assert output_text == expected_text[counter]
      assert output_token_id == expected_token_ids[counter]
      counter += 1
    driver.stop()
    print("Orchestrator driver stopped.")

  @parameterized.expand([True, False])
  async def test_orchestrator_client_tokenization(self, interleaved_mode: bool):
    """Test the multithreaded orchestration."""
    driver = self._setup_driver(interleaved_mode)
    client = orchestrator.LLMOrchestrator(driver=driver)

    # The token ids of  string "AB", [2] will be prepend
    # as BOS.
    token_ids = [65, 66]

    request = jetstream_pb2.DecodeRequest(
        token_content=jetstream_pb2.DecodeRequest.TokenContent(
            token_ids=token_ids
        ),
        max_tokens=3,
    )
    iterator = client.Decode(request)
    # Return token ids only when in client side tokenization mode.
    expected_text = ["", "", "", ""]
    expected_token_ids = [266, 332, 415, None]
    counter = 0
    async for resp in iterator:
      output_text = resp.stream_content.samples[0].text
      token_ids = resp.stream_content.samples[0].token_ids
      output_token_id = token_ids[0] if len(token_ids) > 0 else None
      print(f"actual output: {output_text=} {output_token_id=}")
      assert output_text == expected_text[counter]
      assert output_token_id == expected_token_ids[counter]
      counter += 1
    driver.stop()
    print("Orchestrator driver stopped.")

  @parameterized.expand([True, False])
  def test_should_buffer_response(self, interleaved_mode: bool):
    driver = self._setup_driver(interleaved_mode)
    client = orchestrator.LLMOrchestrator(driver=driver)
    self.assertTrue(
        client.should_buffer_response(
            [ReturnSample(text=["<0xAB>"], token_ids=[13])]
        )
    )
    driver.stop()
    print("Orchestrator driver stopped.")



#  Copyright 2025 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

""" Implementation of the kvcache. """

from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2.flax import aqt_flax
from typing import Any, Optional
from flax import linen as nn
import jax
import jax.numpy as jnp
import common_types


Array = common_types.Array
AxisNames = common_types.AxisNames
AxisIdxes = common_types.AxisIdxes
Config = common_types.Config
KVTensor = aqt_tensor.QTensor

MAX_INT8 = 127.5
MAX_INT4 = 7.5
E4M3_MAX = jnp.finfo(jnp.float8_e4m3fn).max.astype(jnp.float32)

CACHE_BATCH_PREFILL = common_types.CACHE_BATCH_PREFILL
CACHE_BATCH = common_types.CACHE_BATCH
CACHE_SEQUENCE = common_types.CACHE_SEQUENCE
CACHE_HEADS = common_types.CACHE_HEADS
CACHE_KV = common_types.CACHE_KV
CACHE_SCALE_BATCH = common_types.CACHE_SCALE_BATCH
CACHE_SCALE_SEQUENCE = common_types.CACHE_SCALE_SEQUENCE
CACHE_SCALE_HEADS = common_types.CACHE_SCALE_HEADS
CACHE_SCALE_KV = common_types.CACHE_SCALE_KV


def reverse_transpose(transposed_array, transpose_axis_order):
  return jax.numpy.moveaxis(transposed_array, (0, 1, 2, 3), transpose_axis_order)


def transpose_tuple(items: tuple[Any, Any, Any, Any], axis_order: AxisIdxes) -> tuple[Any, Any, Any, Any]:
  return tuple([items[i] for i in axis_order])


class KVQuant:
  """Class to configure quantization for KV cache."""

  axis_cfg = ""
  dtype = None

  def __init__(self, config: Config):
    assert config.quantize_kvcache
    self.axis_cfg = config.kv_quant_axis
    self.dtype = self._get_dtype(config.kv_quant_dtype)

  def _get_dtype(self, dtype_cfg: str):
    if dtype_cfg == "int4":
      return jnp.int4
    if dtype_cfg == "int8":
      return jnp.int8
    if dtype_cfg == "fp8":
      return jnp.float8_e4m3fn
    raise ValueError(f"Invalid kv_quant_dtype: {dtype_cfg}")

  def _get_max_axis(self, axis_names: AxisNames):
    if self.axis_cfg == "dkv":
      return axis_names.index(CACHE_KV)
    if self.axis_cfg == "heads_and_dkv":
      return (axis_names.index(CACHE_HEADS), axis_names.index(CACHE_KV))
    raise ValueError(f"Invalid KV quant axis cfg: {self.axis_cfg}")

  def quantize(self, kv: Array, axis_names: AxisNames):
    """Quantize key/values stored in kvcache."""
    assert self.axis_cfg, "KV quant axis cannot be None"
    max_axis = self._get_max_axis(axis_names)
    scale = jnp.max(jnp.abs(kv), axis=max_axis, keepdims=True)
    if self.dtype == jnp.int8:
      value = jnp.int8(jnp.rint(kv * (MAX_INT8 / scale)))
      return value, scale
    if self.dtype == jnp.int4:
      value = jnp.int4(jnp.rint(kv * (MAX_INT4 / scale)))
      return value, scale
    if self.dtype == jnp.float8_e4m3fn:
      value = jnp.float8_e4m3fn(kv * (E4M3_MAX / scale))
      return value, scale
    raise ValueError(f"Invalid KV quant dtype:{self.dtype}.")

  def einsum_fn_with_rhs_qtensor(
      self,
      kv: Array | aqt_tensor.QTensor,
      rhs_dequant_mode=None,
      rhs_calibration_mode=None,
      lhs_dequant_mode=None,
      lhs_calibration_mode=None,
  ):
    # Assumes kv is already quantized.
    einsum = jnp.einsum
    if isinstance(kv, aqt_tensor.QTensor):
      if kv.qvalue.dtype != jnp.float8_e4m3fn:
        num_bits = 4 if kv.qvalue.dtype == jnp.int4 else 8
        kv_cfg = aqt_config.dot_general_make(
            lhs_bits=None,
            rhs_bits=num_bits,
            bwd_bits=None,
            use_fwd_quant=False,
        )
      else:
        kv_cfg = aqt_config.config_fwd_fp8()

      if rhs_dequant_mode:
        aqt_config.set_fwd_dequant_mode(kv_cfg, rhs_dequant_mode=rhs_dequant_mode)
      if rhs_calibration_mode:
        aqt_config.set_fwd_calibration_mode(
            kv_cfg,
            rhs_calibration_mode=rhs_calibration_mode,
        )
      if lhs_dequant_mode:
        aqt_config.set_fwd_dequant_mode(kv_cfg, lhs_dequant_mode=lhs_dequant_mode)
      if lhs_calibration_mode:
        aqt_config.set_fwd_calibration_mode(
            kv_cfg,
            lhs_calibration_mode=lhs_calibration_mode,
        )
      einsum = aqt_flax.AqtEinsum(
          rhs_quant_mode=aqt_flax.QuantMode.TRAIN,
          lhs_freeze_mode=aqt_flax.FreezerMode.NONE,
          rhs_freeze_mode=aqt_flax.FreezerMode.NONE,
          cfg=kv_cfg,
      )
    return einsum

  def einsum_fn_with_rhs_qtensor_and_dequant(self, value):
    if self.dtype == jnp.float8_e4m3fn:
      return self.einsum_fn_with_rhs_qtensor(
          value,
          lhs_dequant_mode=aqt_config.DequantMode.THIS_INPUT,
          lhs_calibration_mode=aqt_config.CalibrationMode.REMAINING_AXIS,
          rhs_dequant_mode=aqt_config.DequantMode.OTHER_INPUT,
          rhs_calibration_mode=aqt_config.CalibrationMode.REMAINING_AXIS,
      )
    else:
      return self.einsum_fn_with_rhs_qtensor(
          value,
          rhs_dequant_mode=aqt_config.DequantMode.OTHER_INPUT,
          rhs_calibration_mode=aqt_config.CalibrationMode.REMAINING_AXIS,
      )


class KVCache(nn.Module):
  max_prefill_length: int
  max_target_length: int
  dtype: common_types.DType
  kv_quant: Optional[KVQuant] = None
  prefill_cache_logical_axis_names: AxisNames = (CACHE_BATCH_PREFILL, CACHE_SEQUENCE, CACHE_HEADS, CACHE_KV)
  cache_logical_axis_names: AxisNames = (CACHE_BATCH, CACHE_SEQUENCE, CACHE_HEADS, CACHE_KV)
  cache_scale_logical_axis_names: AxisNames = (CACHE_SCALE_BATCH, CACHE_SCALE_SEQUENCE, CACHE_SCALE_HEADS, CACHE_SCALE_KV)
  prefill_cache_axis_order: AxisIdxes = (1, 2, 0, 3)
  ar_cache_axis_order: AxisIdxes = (1, 2, 0, 3)
  use_chunked_prefill: bool = False

  def _get_cached_kv_dtype(self):
    return self.kv_quant.dtype if self.kv_quant else self.dtype

  def _get_cache_scale_logical_shape(self, batch, heads, cache_length):
    assert self.kv_quant
    if self.kv_quant.axis_cfg == "dkv":
      return (batch, cache_length, heads, 1)
    if self.kv_quant.axis_cfg == "heads_and_dkv":
      return (batch, cache_length, 1, 1)
    raise f"Invalid config for kv_quant_axis:{self.kv_quant.axis_cfg}"

  def _get_prefill_cache_vars(self, batch, heads, key_head_size, value_head_size, model_mode):

    cache_length = self.max_prefill_length
    dtype = self._get_cached_kv_dtype()

    if model_mode == common_types.MODEL_MODE_PREFILL:
      cache_logical_axis_names = self.prefill_cache_logical_axis_names
    else:
      cache_logical_axis_names = self.cache_logical_axis_names
    cache_axis_names = transpose_tuple(cache_logical_axis_names, self.prefill_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, key_head_size)
    cache_shape_key = transpose_tuple(cache_logical_shape, self.prefill_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, value_head_size)
    cache_shape_value = transpose_tuple(cache_logical_shape, self.prefill_cache_axis_order)

    cached_key_var = self.variable(
        "cache",
        "cached_prefill_key",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_key,
        dtype,
    )
    cached_value_var = self.variable(
        "cache",
        "cached_prefill_value",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_value,
        dtype,
    )
    if model_mode == common_types.MODEL_MODE_PREFILL:
      segment_id_axis_names = (CACHE_BATCH_PREFILL, CACHE_SEQUENCE)
    else:
      segment_id_axis_names = (CACHE_BATCH, CACHE_SEQUENCE)

    cached_segment_id_var = self.variable(
        "cache",
        "cache_prefill_segment_id",
        nn.with_logical_partitioning(jnp.zeros, segment_id_axis_names),
        (cache_logical_shape[0], cache_length),
        jnp.int32,
    )

    if self.kv_quant:
      cache_scale_logical_shape = self._get_cache_scale_logical_shape(batch, heads, cache_length)
      cache_scale_axis_names = transpose_tuple(self.cache_scale_logical_axis_names, self.prefill_cache_axis_order)
      cache_scale_shape = transpose_tuple(cache_scale_logical_shape, self.prefill_cache_axis_order)

      cached_key_scale_var = self.variable(
          "cache",
          "cached_prefill_key_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
      cached_value_scale_var = self.variable(
          "cache",
          "cached_prefill_value_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
    else:
      cached_key_scale_var = None
      cached_value_scale_var = None

    key_vars = (cached_key_var, cached_key_scale_var)
    value_vars = (cached_value_var, cached_value_scale_var)
    return key_vars, value_vars, cached_segment_id_var

  def _get_ar_cache_vars(self, batch, heads, key_head_size, value_head_size, model_mode):

    dtype = self._get_cached_kv_dtype()
    cache_length = self.max_target_length - self.max_prefill_length

    if model_mode == common_types.MODEL_MODE_PREFILL:
      cache_logical_axis_names = self.prefill_cache_logical_axis_names
    else:
      cache_logical_axis_names = self.cache_logical_axis_names
    cache_axis_names = transpose_tuple(cache_logical_axis_names, self.ar_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, key_head_size)
    cache_shape_key = transpose_tuple(cache_logical_shape, self.ar_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, value_head_size)
    cache_shape_value = transpose_tuple(cache_logical_shape, self.ar_cache_axis_order)

    # TODO(b/339703100): investigate the issue why with_logical_partitioning doesn't enforce sharding
    cached_key_var = self.variable(
        "cache",
        "cached_ar_key",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_key,
        dtype,
    )
    cached_key_var.value = nn.with_logical_constraint(
        cached_key_var.value,
        cache_axis_names,
    )

    cached_value_var = self.variable(
        "cache",
        "cached_ar_value",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_value,
        dtype,
    )
    cached_value_var.value = nn.with_logical_constraint(
        cached_value_var.value,
        cache_axis_names,
    )

    if model_mode == common_types.MODEL_MODE_PREFILL:
      segment_id_axis_names = (CACHE_BATCH_PREFILL, CACHE_SEQUENCE)
    else:
      segment_id_axis_names = (CACHE_BATCH, CACHE_SEQUENCE)
    cached_segment_id_var = self.variable(
        "cache",
        "cache_ar_segment_id",
        nn.with_logical_partitioning(jnp.zeros, segment_id_axis_names),
        (cache_logical_shape[0], cache_length),
        jnp.int32,
    )

    cached_lengths_var = self.variable(
        "cache",
        "cached_ar_lengths",
        nn.with_logical_partitioning(jnp.zeros, (CACHE_BATCH,)),
        (cache_logical_shape[0],),
        jnp.int32,
    )

    if self.kv_quant:
      cache_scale_logical_shape = self._get_cache_scale_logical_shape(batch, heads, cache_length)
      cache_scale_axis_names = transpose_tuple(self.cache_scale_logical_axis_names, self.ar_cache_axis_order)
      cache_scale_shape = transpose_tuple(cache_scale_logical_shape, self.ar_cache_axis_order)

      cached_key_scale_var = self.variable(
          "cache",
          "cached_ar_key_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
      cached_value_scale_var = self.variable(
          "cache",
          "cached_ar_value_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
    else:
      cached_key_scale_var = None
      cached_value_scale_var = None

    cache_index_var = self.variable("cache", "cache_ar_index", nn.with_logical_partitioning(jnp.zeros, ()), (1,), jnp.int32)
    key_vars = (cached_key_var, cached_key_scale_var)
    value_vars = (cached_value_var, cached_value_scale_var)
    return key_vars, value_vars, cached_segment_id_var, cache_index_var, cached_lengths_var

  def chunked_prefill_kv_cache(self, key: Array, value: Array, decoder_segment_ids: Array, previous_chunk: Any = None):
    """
    function returns appropriate prefill_cache if there is previous_chunk already processed
    if no pervious chunk is processed,
    function returns a cache that has non zero first chunk part of key and value

    else
    function updates current key and value at the desired location and returns entire key and value

    """
    batch, _, heads, key_head_size = key.shape
    batch, _, heads, value_head_size = value.shape

    assert key.dtype == value.dtype, "Key and Value Dtypes should match."

    cached_prefill_key_vars, cached_prefill_value_vars, cached_prefill_segment_id_var = self._get_prefill_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )
    # TODO: Find a way to not enable the ar cache for prefill mode.
    _ = self._get_ar_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )  # initialize it now

    key_shaped_for_cache = jnp.transpose(key, self.prefill_cache_axis_order)
    value_shaped_for_cache = jnp.transpose(value, self.prefill_cache_axis_order)

    next_pos = 0
    if previous_chunk != None:
      """
      if there is previous chunk information present,
        1. Fetch the cached key, value
        2. Update current key value at the desired position.
        3. take transpose before returning as that is how the attention op expects the key and value
      """
      next_pos = previous_chunk["true_length_array"].shape[1]
      cached_key = self.get_cached_values(cached_prefill_key_vars, key.dtype, self.prefill_cache_axis_order)
      cached_value = self.get_cached_values(cached_prefill_value_vars, value.dtype, self.prefill_cache_axis_order)
      cached_key_value = jnp.transpose(cached_key, self.prefill_cache_axis_order)
      cached_value_value = jnp.transpose(cached_value, self.prefill_cache_axis_order)

      cached_prefill_key_vars[0].value = jax.lax.dynamic_update_slice(
          cached_key_value, key_shaped_for_cache, (next_pos, 0, 0, 0)
      )

      cached_prefill_value_vars[0].value = jax.lax.dynamic_update_slice(
          cached_value_value, value_shaped_for_cache, (next_pos, 0, 0, 0)
      )
      cached_prefill_segment_id_var.value = decoder_segment_ids
      return (
          jnp.transpose(cached_prefill_key_vars[0].value, self.key_axis_order),
          jnp.transpose(cached_prefill_value_vars[0].value, self.key_axis_order),
          cached_prefill_segment_id_var.value,
      )
    else:
      """
      if there is previous chunk information present,
        1. Fetch the cached key, value
        2. Update current key value at the desired position. (beginning - (0,0,0,0))
        3. take transpose before returning as that is how the attention op expects the key and value

      """
      cached_prefill_key_vars[0].value = jax.lax.dynamic_update_slice(
          cached_prefill_key_vars[0].value, key_shaped_for_cache, (next_pos, 0, 0, 0)
      )
      cached_prefill_value_vars[0].value = jax.lax.dynamic_update_slice(
          cached_prefill_value_vars[0].value, value_shaped_for_cache, (next_pos, 0, 0, 0)
      )
      cached_prefill_segment_id_var.value = decoder_segment_ids
      return (
          jnp.transpose(cached_prefill_key_vars[0].value, self.key_axis_order),
          jnp.transpose(cached_prefill_value_vars[0].value, self.key_axis_order),
          cached_prefill_segment_id_var.value,
      )

  def kv_cache_prefill(
      self,
      key: Array,
      value: Array,
      decoder_segment_ids: Array,
      previous_chunk: Any = None,
  ):
    """In prefill mode, we zero out the existing cache, run the computation and
    prepare the cache as necessary.

    Args:
      key: in shape [b, s, n, d].
      value: in shape [b, s, n, d].
      decoder_segment_ids: [b, s] -- marking segment ids for tokens

    Returns:
      key, value, decoder_segment_id.

    """
    if self.use_chunked_prefill:
      return self.chunked_prefill_kv_cache(key, value, decoder_segment_ids, previous_chunk)

    batch, _, heads, key_head_size = key.shape
    batch, _, heads, value_head_size = value.shape
    assert key.dtype == value.dtype, "Key and Value Dtypes should match."

    cached_prefill_key_vars, cached_prefill_value_vars, cached_prefill_segment_id_var = self._get_prefill_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )
    # TODO: Find a way to not enable the ar cache for prefill mode.
    _ = self._get_ar_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )  # initialize it now

    key_shaped_for_cache = jnp.transpose(key, self.prefill_cache_axis_order)
    value_shaped_for_cache = jnp.transpose(value, self.prefill_cache_axis_order)

    if self.kv_quant:
      prefill_key_axis_names = transpose_tuple(self.cache_logical_axis_names, self.prefill_cache_axis_order)
      key_shaped_for_cache, key_scale_shaped_for_cache = self.kv_quant.quantize(key_shaped_for_cache, prefill_key_axis_names)
      value_shaped_for_cache, value_scale_shaped_for_cache = self.kv_quant.quantize(
          value_shaped_for_cache, prefill_key_axis_names
      )
      cached_prefill_key_vars[1].value = key_scale_shaped_for_cache
      cached_prefill_value_vars[1].value = value_scale_shaped_for_cache

    cached_prefill_key_vars[0].value = key_shaped_for_cache
    cached_prefill_value_vars[0].value = value_shaped_for_cache

    if decoder_segment_ids is not None:
      cached_prefill_segment_id_var.value = decoder_segment_ids
    return key, value, decoder_segment_ids

  def update_ar_key_value(
      self,
      one_token_key: Array,
      one_token_value: Array,
      cached_key_vars: tuple[nn.Variable, nn.Variable | None],
      cached_value_vars: tuple[nn.Variable, nn.Variable | None],
      one_hot_indices: Array,
      lengths: Array,
      use_ragged_attention: bool,
  ) -> None:
    """Adds a single token's results to the ar kv cache

    Args:
        one_token_key (Array): Key of one token to add to the cache
        one_token_value (Array): Value of one token to add to the cache
        cached_ar_key (tuple[nn.Variable, nn.Variable|None],): Cached keys to add new token key to, possibly with scale
        cached_ar_value (tuple[nn.Variable, nn.Variable|None],: Cached values to add new token value to, possible with scale
        one_hot_indices (Array): Location of the new token within the cache

    Returns:
        tuple[Array, Array]: Updated caches for key and value with new token info added
    """

    cached_key_var, cached_key_scale_var = cached_key_vars
    cached_value_var, cached_value_scale_var = cached_value_vars

    # In order to update the key, value caches with the current key and
    # value, we reshape the one_token_key and one_token_value
    one_token_key_shaped_for_cache = jnp.transpose(one_token_key, self.ar_cache_axis_order)
    one_token_value_shaped_for_cache = jnp.transpose(one_token_value, self.ar_cache_axis_order)

    ar_cache_axis_names = transpose_tuple(self.cache_logical_axis_names, self.ar_cache_axis_order)
    if self.kv_quant:
      one_token_key_shaped_for_cache, one_token_key_scale_shaped_for_cache = self.kv_quant.quantize(
          one_token_key_shaped_for_cache, ar_cache_axis_names
      )
      one_token_value_shaped_for_cache, one_token_value_scale_shaped_for_cache = self.kv_quant.quantize(
          one_token_value_shaped_for_cache, ar_cache_axis_names
      )

    ar_cache_update_idx = jnp.squeeze(one_hot_indices)
    ar_cache_sequence_axis = ar_cache_update_axis = ar_cache_axis_names.index(CACHE_SEQUENCE)
    ar_cache_batch_axis = ar_cache_axis_names.index(CACHE_BATCH)


class KVCache(nn.Module):
  """Implementation of the KVCache."""

  max_prefill_length: int
  max_target_length: int
  dtype: common_types.DType
  kv_quant: Optional[KVQuant] = None
  prefill_cache_logical_axis_names: AxisNames = (CACHE_BATCH_PREFILL, CACHE_SEQUENCE, CACHE_HEADS, CACHE_KV)
  cache_logical_axis_names: AxisNames = (CACHE_BATCH, CACHE_SEQUENCE, CACHE_HEADS, CACHE_KV)
  cache_scale_logical_axis_names: AxisNames = (CACHE_SCALE_BATCH, CACHE_SCALE_SEQUENCE, CACHE_SCALE_HEADS, CACHE_SCALE_KV)
  prefill_cache_axis_order: AxisIdxes = (1, 2, 0, 3)
  ar_cache_axis_order: AxisIdxes = (1, 2, 0, 3)
  use_chunked_prefill: bool = False

  def _get_cached_kv_dtype(self):
    return self.kv_quant.dtype if self.kv_quant else self.dtype

  def _get_cache_scale_logical_shape(self, batch, heads, cache_length):
    assert self.kv_quant
    if self.kv_quant.axis_cfg == "dkv":
      return (batch, cache_length, heads, 1)
    if self.kv_quant.axis_cfg == "heads_and_dkv":
      return (batch, cache_length, 1, 1)
    raise f"Invalid config for kv_quant_axis:{self.kv_quant.axis_cfg}"

  def _get_prefill_cache_vars(self, batch, heads, key_head_size, value_head_size, model_mode):

    cache_length = self.max_prefill_length
    dtype = self._get_cached_kv_dtype()

    if model_mode == common_types.MODEL_MODE_PREFILL:
      cache_logical_axis_names = self.prefill_cache_logical_axis_names
    else:
      cache_logical_axis_names = self.cache_logical_axis_names
    cache_axis_names = transpose_tuple(cache_logical_axis_names, self.prefill_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, key_head_size)
    cache_shape_key = transpose_tuple(cache_logical_shape, self.prefill_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, value_head_size)
    cache_shape_value = transpose_tuple(cache_logical_shape, self.prefill_cache_axis_order)

    cached_key_var = self.variable(
        "cache",
        "cached_prefill_key",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_key,
        dtype,
    )
    cached_value_var = self.variable(
        "cache",
        "cached_prefill_value",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_value,
        dtype,
    )
    if model_mode == common_types.MODEL_MODE_PREFILL:
      segment_id_axis_names = (CACHE_BATCH_PREFILL, CACHE_SEQUENCE)
    else:
      segment_id_axis_names = (CACHE_BATCH, CACHE_SEQUENCE)

    cached_segment_id_var = self.variable(
        "cache",
        "cache_prefill_segment_id",
        nn.with_logical_partitioning(jnp.zeros, segment_id_axis_names),
        (cache_logical_shape[0], cache_length),
        jnp.int32,
    )

    if self.kv_quant:
      cache_scale_logical_shape = self._get_cache_scale_logical_shape(batch, heads, cache_length)
      cache_scale_axis_names = transpose_tuple(self.cache_scale_logical_axis_names, self.prefill_cache_axis_order)
      cache_scale_shape = transpose_tuple(cache_scale_logical_shape, self.prefill_cache_axis_order)

      cached_key_scale_var = self.variable(
          "cache",
          "cached_prefill_key_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
      cached_value_scale_var = self.variable(
          "cache",
          "cached_prefill_value_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
    else:
      cached_key_scale_var = None
      cached_value_scale_var = None

    key_vars = (cached_key_var, cached_key_scale_var)
    value_vars = (cached_value_var, cached_value_scale_var)
    return key_vars, value_vars, cached_segment_id_var

  def _get_ar_cache_vars(self, batch, heads, key_head_size, value_head_size, model_mode):

    dtype = self._get_cached_kv_dtype()
    cache_length = self.max_target_length - self.max_prefill_length

    if model_mode == common_types.MODEL_MODE_PREFILL:
      cache_logical_axis_names = self.prefill_cache_logical_axis_names
    else:
      cache_logical_axis_names = self.cache_logical_axis_names
    cache_axis_names = transpose_tuple(cache_logical_axis_names, self.ar_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, key_head_size)
    cache_shape_key = transpose_tuple(cache_logical_shape, self.ar_cache_axis_order)

    cache_logical_shape = (batch, cache_length, heads, value_head_size)
    cache_shape_value = transpose_tuple(cache_logical_shape, self.ar_cache_axis_order)

    # TODO(b/339703100): investigate the issue why with_logical_partitioning doesn't enforce sharding
    cached_key_var = self.variable(
        "cache",
        "cached_ar_key",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_key,
        dtype,
    )
    cached_key_var.value = nn.with_logical_constraint(
        cached_key_var.value,
        cache_axis_names,
    )

    cached_value_var = self.variable(
        "cache",
        "cached_ar_value",
        nn.with_logical_partitioning(jnp.zeros, cache_axis_names),
        cache_shape_value,
        dtype,
    )
    cached_value_var.value = nn.with_logical_constraint(
        cached_value_var.value,
        cache_axis_names,
    )

    if model_mode == common_types.MODEL_MODE_PREFILL:
      segment_id_axis_names = (CACHE_BATCH_PREFILL, CACHE_SEQUENCE)
    else:
      segment_id_axis_names = (CACHE_BATCH, CACHE_SEQUENCE)
    cached_segment_id_var = self.variable(
        "cache",
        "cache_ar_segment_id",
        nn.with_logical_partitioning(jnp.zeros, segment_id_axis_names),
        (cache_logical_shape[0], cache_length),
        jnp.int32,
    )

    cached_lengths_var = self.variable(
        "cache",
        "cached_ar_lengths",
        nn.with_logical_partitioning(jnp.zeros, (CACHE_BATCH,)),
        (cache_logical_shape[0],),
        jnp.int32,
    )

    if self.kv_quant:
      cache_scale_logical_shape = self._get_cache_scale_logical_shape(batch, heads, cache_length)
      cache_scale_axis_names = transpose_tuple(self.cache_scale_logical_axis_names, self.ar_cache_axis_order)
      cache_scale_shape = transpose_tuple(cache_scale_logical_shape, self.ar_cache_axis_order)

      cached_key_scale_var = self.variable(
          "cache",
          "cached_ar_key_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
      cached_value_scale_var = self.variable(
          "cache",
          "cached_ar_value_scale",
          nn.with_logical_partitioning(jnp.zeros, cache_scale_axis_names),
          cache_scale_shape,
          jnp.bfloat16,
      )
    else:
      cached_key_scale_var = None
      cached_value_scale_var = None

    cache_index_var = self.variable("cache", "cache_ar_index", nn.with_logical_partitioning(jnp.zeros, ()), (1,), jnp.int32)
    key_vars = (cached_key_var, cached_key_scale_var)
    value_vars = (cached_value_var, cached_value_scale_var)
    return key_vars, value_vars, cached_segment_id_var, cache_index_var, cached_lengths_var

  def chunked_prefill_kv_cache(self, key: Array, value: Array, decoder_segment_ids: Array, previous_chunk: Any = None):
    """
    function returns appropriate prefill_cache if there is previous_chunk already processed
    if no pervious chunk is processed,
    function returns a cache that has non zero first chunk part of key and value

    else
    function updates current key and value at the desired location and returns entire key and value

    """
    batch, _, heads, key_head_size = key.shape
    batch, _, heads, value_head_size = value.shape

    assert key.dtype == value.dtype, "Key and Value Dtypes should match."

    cached_prefill_key_vars, cached_prefill_value_vars, cached_prefill_segment_id_var = self._get_prefill_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )
    # TODO: Find a way to not enable the ar cache for prefill mode.
    _ = self._get_ar_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )  # initialize it now

    key_shaped_for_cache = jnp.transpose(key, self.prefill_cache_axis_order)
    value_shaped_for_cache = jnp.transpose(value, self.prefill_cache_axis_order)

    next_pos = 0
    if previous_chunk != None:
      """
      if there is previous chunk information present,
        1. Fetch the cached key, value
        2. Update current key value at the desired position.
        3. take transpose before returning as that is how the attention op expects the key and value
      """
      next_pos = previous_chunk["true_length_array"].shape[1]
      cached_key = self.get_cached_values(cached_prefill_key_vars, key.dtype, self.prefill_cache_axis_order)
      cached_value = self.get_cached_values(cached_prefill_value_vars, value.dtype, self.prefill_cache_axis_order)
      cached_key_value = jnp.transpose(cached_key, self.prefill_cache_axis_order)
      cached_value_value = jnp.transpose(cached_value, self.prefill_cache_axis_order)

      cached_prefill_key_vars[0].value = jax.lax.dynamic_update_slice(
          cached_key_value, key_shaped_for_cache, (next_pos, 0, 0, 0)
      )

      cached_prefill_value_vars[0].value = jax.lax.dynamic_update_slice(
          cached_value_value, value_shaped_for_cache, (next_pos, 0, 0, 0)
      )
      cached_prefill_segment_id_var.value = decoder_segment_ids
      return (
          jnp.transpose(cached_prefill_key_vars[0].value, self.key_axis_order),
          jnp.transpose(cached_prefill_value_vars[0].value, self.key_axis_order),
          cached_prefill_segment_id_var.value,
      )
    else:
      """
      if there is previous chunk information present,
        1. Fetch the cached key, value
        2. Update current key value at the desired position. (beginning - (0,0,0,0))
        3. take transpose before returning as that is how the attention op expects the key and value

      """
      cached_prefill_key_vars[0].value = jax.lax.dynamic_update_slice(
          cached_prefill_key_vars[0].value, key_shaped_for_cache, (next_pos, 0, 0, 0)
      )
      cached_prefill_value_vars[0].value = jax.lax.dynamic_update_slice(
          cached_prefill_value_vars[0].value, value_shaped_for_cache, (next_pos, 0, 0, 0)
      )
      cached_prefill_segment_id_var.value = decoder_segment_ids
      return (
          jnp.transpose(cached_prefill_key_vars[0].value, self.key_axis_order),
          jnp.transpose(cached_prefill_value_vars[0].value, self.key_axis_order),
          cached_prefill_segment_id_var.value,
      )

  def kv_cache_prefill(
      self,
      key: Array,
      value: Array,
      decoder_segment_ids: Array,
      previous_chunk: Any = None,
  ):
    """In prefill mode, we zero out the existing cache, run the computation and
    prepare the cache as necessary.

    Args:
      key: in shape [b, s, n, d].
      value: in shape [b, s, n, d].
      decoder_segment_ids: [b, s] -- marking segment ids for tokens

    Returns:
      key, value, decoder_segment_id.

    """
    if self.use_chunked_prefill:
      return self.chunked_prefill_kv_cache(key, value, decoder_segment_ids, previous_chunk)

    batch, _, heads, key_head_size = key.shape
    batch, _, heads, value_head_size = value.shape
    assert key.dtype == value.dtype, "Key and Value Dtypes should match."

    cached_prefill_key_vars, cached_prefill_value_vars, cached_prefill_segment_id_var = self._get_prefill_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )
    # TODO: Find a way to not enable the ar cache for prefill mode.
    _ = self._get_ar_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_PREFILL
    )  # initialize it now

    key_shaped_for_cache = jnp.transpose(key, self.prefill_cache_axis_order)
    value_shaped_for_cache = jnp.transpose(value, self.prefill_cache_axis_order)

    if self.kv_quant:
      prefill_key_axis_names = transpose_tuple(self.cache_logical_axis_names, self.prefill_cache_axis_order)
      key_shaped_for_cache, key_scale_shaped_for_cache = self.kv_quant.quantize(key_shaped_for_cache, prefill_key_axis_names)
      value_shaped_for_cache, value_scale_shaped_for_cache = self.kv_quant.quantize(
          value_shaped_for_cache, prefill_key_axis_names
      )
      cached_prefill_key_vars[1].value = key_scale_shaped_for_cache
      cached_prefill_value_vars[1].value = value_scale_shaped_for_cache

    cached_prefill_key_vars[0].value = key_shaped_for_cache
    cached_prefill_value_vars[0].value = value_shaped_for_cache

    if decoder_segment_ids is not None:
      cached_prefill_segment_id_var.value = decoder_segment_ids
    return key, value, decoder_segment_ids

  def update_ar_key_value(
      self,
      one_token_key: Array,
      one_token_value: Array,
      cached_key_vars: tuple[nn.Variable, nn.Variable | None],
      cached_value_vars: tuple[nn.Variable, nn.Variable | None],
      one_hot_indices: Array,
      lengths: Array,
      use_ragged_attention: bool,
  ) -> None:
    """Adds a single token's results to the ar kv cache

    Args:
        one_token_key (Array): Key of one token to add to the cache
        one_token_value (Array): Value of one token to add to the cache
        cached_ar_key (tuple[nn.Variable, nn.Variable|None],): Cached keys to add new token key to, possibly with scale
        cached_ar_value (tuple[nn.Variable, nn.Variable|None],: Cached values to add new token value to, possible with scale
        one_hot_indices (Array): Location of the new token within the cache

    Returns:
        tuple[Array, Array]: Updated caches for key and value with new token info added
    """

    cached_key_var, cached_key_scale_var = cached_key_vars
    cached_value_var, cached_value_scale_var = cached_value_vars

    # In order to update the key, value caches with the current key and
    # value, we reshape the one_token_key and one_token_value
    one_token_key_shaped_for_cache = jnp.transpose(one_token_key, self.ar_cache_axis_order)
    one_token_value_shaped_for_cache = jnp.transpose(one_token_value, self.ar_cache_axis_order)

    ar_cache_axis_names = transpose_tuple(self.cache_logical_axis_names, self.ar_cache_axis_order)
    if self.kv_quant:
      one_token_key_shaped_for_cache, one_token_key_scale_shaped_for_cache = self.kv_quant.quantize(
          one_token_key_shaped_for_cache, ar_cache_axis_names
      )
      one_token_value_shaped_for_cache, one_token_value_scale_shaped_for_cache = self.kv_quant.quantize(
          one_token_value_shaped_for_cache, ar_cache_axis_names
      )

    ar_cache_update_idx = jnp.squeeze(one_hot_indices)
    ar_cache_sequence_axis = ar_cache_update_axis = ar_cache_axis_names.index(CACHE_SEQUENCE)
    ar_cache_batch_axis = ar_cache_axis_names.index(CACHE_BATCH)

    if use_ragged_attention:
      cache_locations = [slice(None)] * 4
      new_token_locations = [slice(None)] * 4
      new_token_locations[ar_cache_sequence_axis] = 0

      def key_body(i, val):
        cache_locations[ar_cache_batch_axis] = i
        cache_locations[ar_cache_sequence_axis] = lengths[i]
        new_token_locations[ar_cache_batch_axis] = i
        return val.at[tuple(cache_locations)].set(one_token_key_shaped_for_cache[tuple(new_token_locations)])

      def value_body(i, val):
        cache_locations[ar_cache_batch_axis] = i
        cache_locations[ar_cache_sequence_axis] = lengths[i]
        new_token_locations[ar_cache_batch_axis] = i
        return val.at[tuple(cache_locations)].set(one_token_value_shaped_for_cache[tuple(new_token_locations)])

      cached_key_var.value = jax.lax.fori_loop(
          0, one_token_key_shaped_for_cache.shape[0], key_body, cached_key_var.value, unroll=8
      )
      cached_value_var.value = jax.lax.fori_loop(
          0, one_token_value_shaped_for_cache.shape[0], value_body, cached_value_var.value, unroll=8
      )

    else:
      one_hot_indices = one_hot_indices.astype(int)
      cached_key_var.value = jax.lax.dynamic_update_index_in_dim(
          cached_key_var.value, one_token_key_shaped_for_cache, ar_cache_update_idx, ar_cache_update_axis
      )
      cached_value_var.value = jax.lax.dynamic_update_index_in_dim(
          cached_value_var.value, one_token_value_shaped_for_cache, ar_cache_update_idx, ar_cache_update_axis
      )
    cached_key_var.value = nn.with_logical_constraint(cached_key_var.value, ar_cache_axis_names)
    cached_value_var.value = nn.with_logical_constraint(cached_value_var.value, ar_cache_axis_names)

    if self.kv_quant:
      ar_cache_scale_axis_names = transpose_tuple(self.cache_scale_logical_axis_names, self.ar_cache_axis_order)
      ar_cache_scale_update_axis = ar_cache_scale_axis_names.index(CACHE_SCALE_SEQUENCE)
      assert cached_key_scale_var is not None, "cached_key_scale_var cannot be None"
      assert cached_value_scale_var is not None, "cached_value_scale_var cannot be None"
      cached_key_scale_var.value = jax.lax.dynamic_update_index_in_dim(
          cached_key_scale_var.value, one_token_key_scale_shaped_for_cache, ar_cache_update_idx, ar_cache_scale_update_axis
      )
      cached_value_scale_var.value = jax.lax.dynamic_update_index_in_dim(
          cached_value_scale_var.value,
          one_token_value_scale_shaped_for_cache,
          ar_cache_update_idx,
          ar_cache_scale_update_axis,
      )
    return

  def get_cached_values(self, cache_vars, target_dtype, cache_axis_order) -> jax.Array | KVTensor:
    cache_var, cache_scale_var = cache_vars
    cache_value = cache_var.value
    if cache_scale_var is not None:
      scale_value = cache_scale_var.value
      dtype = cache_value.dtype
      if dtype == jnp.int8:
        scale_value /= MAX_INT8
      elif dtype == jnp.int4:
        scale_value /= MAX_INT4
      elif dtype == jnp.float8_e4m3fn:
        scale_value /= E4M3_MAX

      cache_value = KVTensor(qvalue=cache_value, scale=[scale_value], scale_t=None, dequant_dtype=target_dtype, bias=[])
    cache_value_in_logical_shape = jax.tree.map(lambda x: reverse_transpose(x, cache_axis_order), cache_value)
    return cache_value_in_logical_shape

  def kv_cache_autoregressive(
      self,
      key: Array,
      value: Array,
      use_ragged_attention: bool = False,
  ):
    """In autoregressive mode, we update the cache for this entry and
       then return the full cache.

    Args:
      key: in shape [b, 1, n, d].
      value: in shape [b, 1, n, d].
      decoder_segment_ids: [b, 1] -- marking segment ids for tokens

    Returns:
      tuple of (key, value, segment_id) for both prefill and ar cache,
    Raises:
      ValueError: when key/value shape is not [batch, 1, num_heads, heads_dim].
    """
    batch, sequence, heads, key_head_size = key.shape
    batch, sequence, heads, value_head_size = value.shape

    if sequence != 1:
      raise ValueError(f"Sequence length should be 1 during autoregression, got {sequence=}")

    cached_ar_key_vars, cached_ar_value_vars, cached_ar_segment_id_var, cache_ar_index_var, cache_ar_lengths_var = (
        self._get_ar_cache_vars(batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_AUTOREGRESSIVE)
    )

    self.update_ar_key_value(
        key,
        value,
        cached_ar_key_vars,
        cached_ar_value_vars,
        cache_ar_index_var.value,
        cache_ar_lengths_var.value,
        use_ragged_attention,
    )
    active_indicator = jnp.zeros((batch, 1), dtype=jnp.int32) + common_types.DECODING_ACTIVE_SEQUENCE_INDICATOR
    cached_ar_segment_id_var.value = jax.lax.dynamic_update_index_in_dim(
        cached_ar_segment_id_var.value, active_indicator, jnp.squeeze(cache_ar_index_var.value), 1
    )
    cache_ar_index_var.value = jnp.mod(cache_ar_index_var.value + 1, self.max_target_length - self.max_prefill_length)
    cache_ar_lengths_var.value = cache_ar_lengths_var.value.at[:].add(1)

    # The below retrieves the existing prefill cache variables, not creating new ones
    cached_prefill_key_vars, cached_prefill_value_vars, cached_prefill_segment_id_var = self._get_prefill_cache_vars(
        batch, heads, key_head_size, value_head_size, common_types.MODEL_MODE_AUTOREGRESSIVE
    )

    cached_prefill = (
        self.get_cached_values(cached_prefill_key_vars, key.dtype, self.prefill_cache_axis_order),
        self.get_cached_values(cached_prefill_value_vars, value.dtype, self.prefill_cache_axis_order),
        cached_prefill_segment_id_var.value,
    )

    cached_ar = (
        self.get_cached_values(cached_ar_key_vars, key.dtype, self.ar_cache_axis_order),
        self.get_cached_values(cached_ar_value_vars, value.dtype, self.ar_cache_axis_order),
        cached_ar_segment_id_var.value,
        cache_ar_lengths_var.value,
    )
    return cached_prefill, cached_ar

  @nn.compact
  def __call__(
      self,
      key: Array,
      value: Array,
      decoder_segment_ids: Array,
      model_mode: str,
      use_ragged_attention: bool = False,
      previous_chunk: Any = None,
  ) -> tuple:
    """KV cache takes the current state and updates the state accordingly.

    The key and value have dimension [b, s, n_kv, d],
    but we cache them with a reshape as defined in *_axis_order config as a TPU
    fusion optimization. This also enables the "scatter via one-hot
    broadcast" trick, which means we do a one-hot broadcast instead of a
    scatter/gather operations, resulting in a 3-4x speedup in practice.

    Args:
      key: in shape [b, s, n_kv, d].
      value: in shape [b, s, n_kv, d].
      model_mode: model mode controlling model

    Returns:
      two tuples of (k, v, decoder_segments) -- either can be Nones

    """
    if model_mode == common_types.MODEL_MODE_PREFILL:
      return self.kv_cache_prefill(key, value, decoder_segment_ids, previous_chunk), None
    elif model_mode == common_types.MODEL_MODE_AUTOREGRESSIVE:
      return self.kv_cache_autoregressive(key, value, use_ragged_attention)
    else:
      raise ValueError(f"Model Mode isn't supported! {model_mode=}")



import argparse
from transformers import AutoTokenizer
import nltk
import evaluate
import numpy as np
import json


def get_args():
  parser = argparse.ArgumentParser()
  parser.add_argument(
      "--checkpoint-path",
      required=True,
      help="Path to Llama2-70b-hf-chat checkpoint",
  )
  parser.add_argument(
      "--mlperf-accuracy-file",
      required=True,
      help="path to mlperf_log_accuracy.json",
  )
  parser.add_argument(
      "--dataset-file",
      required=True,
      help="path to processed openorca validation set",
  )
  parser.add_argument("--verbose", action="store_true", help="verbose messages")
  parser.add_argument(
      "--dtype",
      default="int64",
      help="dtype of the accuracy log",
      choices=["int32", "int64", "float"],
  )
  args = parser.parse_args()
  return args


def get_groundtruth(processed_dataset_file):
  import pandas as pd

  data = pd.read_pickle(processed_dataset_file)
  ground_truths = data["output"]
  return ground_truths


def postprocess_text(preds, targets):
  preds = [pred.strip() for pred in preds]
  targets = [target.strip() for target in targets]

  # rougeLSum expects newline after each sentence
  preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in preds]
  targets = ["\n".join(nltk.sent_tokenize(target)) for target in targets]

  return preds, targets


def main():

  args = get_args()
  dataset_path = args.dataset_file
  checkpoint_path = args.checkpoint_path
  metric = evaluate.load("rouge")
  nltk.download("punkt")

  tokenizer = AutoTokenizer.from_pretrained(
      checkpoint_path,
      model_max_length=2048,
      padding_side="left",
      use_fast=False,
  )

  targets = get_groundtruth(args.dataset_file)

  target_required = []
  preds_token_ids = []

  eval_dtype = np.int64
  if args.dtype == "int32":
    eval_dtype = np.int32
  elif args.dtype == "float":
    eval_dtype = np.float32

  with open(args.mlperf_accuracy_file, "r") as f:
    results = json.load(f)

  seen = set()
  gen_tok_len = 0
  for pred in results:
    qsl_idx = pred["qsl_idx"]
    if qsl_idx in seen:
      continue

    seen.add(qsl_idx)
    target = targets[qsl_idx]
    target_required.append(target)
    pred = np.frombuffer(bytes.fromhex(pred["data"]), eval_dtype)

    gen_tok_len += len(pred)
    preds_token_ids.append(pred)

  preds_decoded_text = tokenizer.batch_decode(
      preds_token_ids, skip_special_tokens=True
  )

  preds, targets = postprocess_text(preds_decoded_text, target_required)

  result = metric.compute(
      predictions=preds,
      references=targets,
      use_stemmer=True,
      use_aggregator=False,
  )
  result = {k: round(np.mean(v) * 100, 4) for k, v in result.items()}
  prediction_lens = [len(pred) for pred in preds]
  gen_num = len(preds)

  result = {
      **result,
      "gen_len": np.sum(prediction_lens),
      "gen_num": gen_num,
      "gen_tok_len": gen_tok_len,
      "tokens_per_sample": round(gen_tok_len / gen_num, 1),
  }

  print("\nResults\n")
  print(result)


if __name__ == "__main__":
  main()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Norm Module"""

import jax
from jax import numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P
from inference.nn import Module, Parameter
from inference import parallel


class RMSNorm(Module):

  def __init__(self, dim, eps, parallel_config: parallel.RMSNormParallelConfig):
    super().__init__()
    self.dim = dim
    self.eps = eps
    self.parallel_config = parallel_config

    self.weight = Parameter(jnp.zeros((dim,)))
    mesh = parallel_config.mesh
    if parallel_config.activation_sharded:
      self.weight.sharding = NamedSharding(mesh, P(parallel.tp_axis_names()))
    else:
      self.weight.sharding = NamedSharding(mesh, P(None))

  def __call__(self, input):
    input_dtype = input.dtype
    input = input.astype(jnp.float32)
    mean_square = jnp.mean(jax.lax.square(input), axis=-1, keepdims=True)

    if self.parallel_config.activation_sharded:
      axis_names = parallel.tp_axis_names()
      mean_square = parallel.ops.all_reduce(
          mean_square, axis_names
      ) / parallel.get_num_partitions(axis_names)

    mean_square = jax.lax.rsqrt(mean_square + self.eps)
    input *= mean_square
    return self.weight * input.astype(input_dtype)



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Usage:

python3 MaxText/scratch_code/generate_hf_golden_logits.py --model-id=deepseek-ai/DeepSeek-V2-Lite --output-path=golden_DeepSeek-V2-Lite.jsonl --prompts='I love to,Today is a,What is the', --gcs_bucket=my-gcs-bucket

"""

import os
import torch
import argparse
from transformers import AutoTokenizer, AutoModelForCausalLM
import jsonlines
from google.cloud import storage

# Load the tokenizer and model from Hugging Face


def upload_blob(bucket_name, source_file_name, destination_blob_name):
  """Uploads a file to the bucket."""
  storage_client = storage.Client()
  bucket = storage_client.get_bucket(bucket_name)
  blob = bucket.blob(destination_blob_name)
  blob.upload_from_filename(source_file_name)


def save_golden_logits(model_id, output_path, prompt_texts, gcs_bucket):
  tokenizer = AutoTokenizer.from_pretrained(model_id)
  model = AutoModelForCausalLM.from_pretrained(
      model_id,
      torch_dtype=torch.float32,
  )

  all_data_to_save = []
  for prompt_text in prompt_texts:
    # Encode the prompt text
    input_ids = tokenizer.encode(prompt_text, return_tensors="pt")

    # Get the logits for the prompt + completion
    with torch.no_grad():
      outputs = model(input_ids)
      logits = outputs.logits.cpu().numpy().astype("float32")

      # Prepare data to be saved
      data_to_save = {
          "prompt": prompt_text,
          "tokens": input_ids.tolist()[0],
          "logits": logits.tolist()[0],  # Convert numpy array to list for JSON serialization
      }
      all_data_to_save.append(data_to_save)

  with jsonlines.open(output_path, "w") as f:
    f.write_all(all_data_to_save)
    print(f"File is stored locally at {output_path}.")

  if gcs_bucket:
    upload_blob(gcs_bucket, output_path, f"golden-logits/{model_id}/{output_path}")
    print(f"File is uploaded to gs://{gcs_bucket}/golden-logits/{model_id}/{output_path}.")


def main(raw_args=None) -> None:
  parser = argparse.ArgumentParser()
  parser.add_argument("--model-id", type=str, required=True, help="The identifier of the model to use.")
  parser.add_argument("--output-path", type=str, required=True, help="The path to save the generated golden logits.")
  parser.add_argument("--prompts", type=str, required=True, help="A comma-separated list of prompts.")
  parser.add_argument(
      "--gcs-bucket", type=str, required=False, default=None, help="A GCS bucket to store logits, without gs://."
  )
  args = parser.parse_args(raw_args)
  prompts = [prompt for prompt in args.prompts.split(",")]
  save_golden_logits(args.model_id, args.output_path, prompts, args.gcs_bucket)


if __name__ == "__main__":
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluate accuracy of JetStream online serving."""

import argparse
import nltk
import evaluate
import json
import re

import numpy as np
from benchmarks.math_utils import extract_numbers, post_processing_math_ans, sympify_set


def extract_boxed_answers(text):
  pieces = text.split("boxed{")
  if len(pieces) == 1:
    return [""]
  piece = pieces[1]
  ans = []
  for piece in pieces[1:]:
    n = 0
    for i in range(len(piece)):
      if piece[i] == "{":
        n += 1
      elif piece[i] == "}":
        n -= 1
        if n < 0:
          if i + 1 < len(piece) and piece[i + 1] == "%":
            ans.append(piece[: i + 1])
            break
          else:
            ans.append(piece[:i])
            break
  if ans:
    return ans
  else:
    return [""]


def extract_answer(pred_str, exhaust=False):
  pred = []
  if "boxed{" in pred_str:
    pred = extract_boxed_answers(pred_str)
  elif "Answer:" in pred_str:
    matches = re.findall(r"Answer:[\*]*\s+(\S*.*)", pred_str)
    if matches:
      pred = [extract_numbers(matches[-1])]
  elif "the answer is" in pred_str:
    pred = [extract_numbers(pred_str.split("the answer is")[-1].strip())]
  elif "final answer is $" in pred_str and "$. I hope" in pred_str:
    tmp = pred_str.split("final answer is $", 1)[1]
    pred = [tmp.split("$. I hope", 1)[0].strip()]
  else:  # use the last number
    pattern = r"-?\d*\.?\d+"
    ans = re.findall(pattern, pred_str.replace(",", ""))
    if len(ans) >= 1:
      ans = ans[-1]
    else:
      ans = ""
    if ans:
      pred.append(ans)
  # multiple line
  pred_list = []
  for ans in pred:
    ans = ans.replace("<|end_of_text|>", "")
    ans = ans.strip().split("\n")[0]
    ans = ans.lstrip(":")
    ans = ans.lstrip("$")
    ans = ans.rstrip("$")
    ans = ans.rstrip(".")
    ans = ans.rstrip("/")
    pred_list.append(ans)
  if exhaust:
    return pred_list
  else:
    return pred_list[-1] if pred_list else ""


def postprocess_text(preds, targets):
  preds = [pred.strip() for pred in preds]
  targets = [target.strip() for target in targets]

  # rougeLSum expects newline after each sentence
  preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in preds]
  targets = ["\n".join(nltk.sent_tokenize(target)) for target in targets]

  return preds, targets


def eval_accuracy(request_outputs_dict, match_type):
  preds = []
  targets = []
  for output in request_outputs_dict:
    preds.append(output["generated_text"])
    targets.append(output["original_output"])

  if match_type == "math":
    correct_ans = 0
    wrong_ans = 0
    for p, t in zip(preds, targets):

      p = extract_answer(p)
      ans_set = post_processing_math_ans(p)
      sympified_ans_set = sympify_set(ans_set)

      target_set = post_processing_math_ans(t)
      sympified_target_set = sympify_set(target_set)

      if sympified_target_set == sympified_ans_set:
        correct_ans += 1
        continue
      wrong_ans += 1
    total_ans = correct_ans + wrong_ans
    result = {}
    result["literal"] = correct_ans / total_ans if total_ans > 0 else 0.0
    result["gen_len"] = total_ans
    result["gen_num"] = total_ans

  else:
    metric = evaluate.load("rouge")
    nltk.download("punkt_tab")
    preds, targets = postprocess_text(preds, targets)
    result = metric.compute(
        predictions=preds,
        references=targets,
        use_stemmer=True,
        use_aggregator=False,
    )
    result = {k: float(round(np.mean(v) * 100, 4)) for k, v in result.items()}
    prediction_lens = [len(pred) for pred in preds]
    result["gen_len"] = int(np.sum(prediction_lens))
    result["gen_num"] = len(preds)

  print("\nResults\n")
  print(result)
  return result


def main(args):
  with open(args.output_path, "r", encoding="utf-8") as f:
    request_outputs_dict = json.load(f)

  eval_accuracy(request_outputs_dict, args.match_type)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument(
      "--output_path",
      type=str,
      default="/tmp/request-outputs.json",
      help="File path which has original_output and inference generated_text.",
  )
  parser.add_argument(
      "--match_type",
      type=str,
      default="rouge",
      nargs="?",
      help="Optional, values are 'rouge' or 'math'. The way to measure the "
      "accuracy of the results. ",
  )

  parsed_args = parser.parse_args()

  main(parsed_args)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""AttentionOps Module"""

import dataclasses
import jax
from jax import numpy as jnp
import jax.experimental
from jax.sharding import NamedSharding
from inference import kernel
from inference.nn.module import Module
from inference.utils import *


@register_flat_dataclass_as_pytree
@dataclasses.dataclass
class KVCache:
  k: jax.Array | NamedSharding
  v: jax.Array | NamedSharding


@register_flat_dataclass_as_pytree
@dataclasses.dataclass
class AttentionMetadata:
  prefill_length: (
      jax.Array | NamedSharding
  )  # shape: []; Prefill True length without padding
  prefill_pos: jax.Array | NamedSharding  # shape: [padded length]
  prefill_page_table: jax.Array | NamedSharding  # shape: [max_len // page_size]

  generate_pos: jax.Array | NamedSharding  # shape: [generate_batch_size]
  generate_page_table: (
      jax.Array | NamedSharding
  )  # shape: [generate_batch, max_len // page_size]


class AttentionOps(Module):

  def __init__(self, num_attn_heads, num_kv_heads, head_dim):
    super().__init__()
    self.num_attn_heads = num_attn_heads
    self.num_kv_heads = num_kv_heads
    self.head_dim = head_dim

  def _write_prefill_kv_to_kv_cache(
      self, k, v, kv_cache: KVCache, unpadded_len, page_table
  ):
    padded_prefill_len = k.shape[0]
    num_kv_heads_per_device = kv_cache.k.shape[0]
    page_size = kv_cache.k.shape[2]
    num_pages = padded_prefill_len // page_size
    num_pages = jnp.where(num_pages < 1, 1, num_pages)
    num_active_pages, reminder = jnp.divmod(unpadded_len, page_size)
    num_active_pages += jnp.where(reminder > 0, 1, 0)

    k = k.transpose((1, 0, 2))
    v = v.transpose((1, 0, 2))
    # kv shape after the change: (num_kv_heads, num_pages, page_size, head_dim)
    k = k.reshape(
        (num_kv_heads_per_device, -1, page_size, self.head_dim)
    ).astype(kv_cache.k.dtype)
    v = v.reshape(
        (num_kv_heads_per_device, -1, page_size, self.head_dim)
    ).astype(kv_cache.v.dtype)

    def update_cond(carry):
      _, idx = carry
      return idx < num_active_pages

    def per_page_update(carry):
      kv_cache, idx = carry
      page_k = k[:, idx, :, :][:, None, :, :]
      page_v = v[:, idx, :, :][:, None, :, :]
      mapped_idx = page_table[idx]
      kv_cache.k = jax.lax.dynamic_update_slice_in_dim(
          kv_cache.k,
          page_k,
          mapped_idx,
          axis=1,
      )
      kv_cache.v = jax.lax.dynamic_update_slice_in_dim(
          kv_cache.v,
          page_v,
          mapped_idx,
          axis=1,
      )
      idx += 1
      return kv_cache, idx

    idx = 0
    kv_cache, idx = jax.lax.while_loop(
        update_cond, per_page_update, (kv_cache, idx)
    )

    return kv_cache

  def _write_generate_kv_to_kv_cache(self, k, v, kv_cache, pos, page_table):
    k = k.transpose((1, 0, 2))
    v = v.transpose((1, 0, 2))

    k = k.astype(kv_cache.k.dtype)
    v = v.astype(kv_cache.v.dtype)

    num_tokens = k.shape[1]
    num_kv_heads_per_device, num_pages, page_size, head_dim = kv_cache.k.shape
    page_idx, offset = jnp.divmod(pos, page_size)
    page_to_update = page_table[jnp.arange(0, num_tokens), page_idx]

    mapped_page_to_update = page_to_update * page_size + offset
    mapped_page_to_update = jnp.tile(
        mapped_page_to_update, num_kv_heads_per_device
    )

    kv_heads_axis_stride = (
        jnp.repeat(jnp.arange(0, num_kv_heads_per_device), num_tokens)
        * num_pages
        * page_size
    )
    mapped_page_to_update = kv_heads_axis_stride + mapped_page_to_update

    k = k.reshape(-1, head_dim)
    v = v.reshape(-1, head_dim)

    kv_cache.k = kv_cache.k.reshape(-1, head_dim)
    kv_cache.v = kv_cache.v.reshape(-1, head_dim)

    kv_cache.k = kv_cache.k.at[mapped_page_to_update, :].set(k)
    kv_cache.v = kv_cache.v.at[mapped_page_to_update, :].set(v)

    kv_cache.k = kv_cache.k.reshape(
        num_kv_heads_per_device, num_pages, page_size, head_dim
    )
    kv_cache.v = kv_cache.v.reshape(
        num_kv_heads_per_device, num_pages, page_size, head_dim
    )

    return kv_cache

  def _prefill(
      self, q, k, v, kv_cache: KVCache, attn_metadata: AttentionMetadata
  ):
    kv_cache = self._write_prefill_kv_to_kv_cache(
        k,
        v,
        kv_cache,
        attn_metadata.prefill_length,
        attn_metadata.prefill_page_table,
    )
    output = kernel.chunked_prefill_attention(
        q,
        kv_cache.k,
        kv_cache.v,
        attn_metadata.prefill_length,
        attn_metadata.prefill_page_table,
    )
    #        if self.num_attn_heads == self.num_kv_heads:
    #            output = kernel.vanilla_prefill_mha(q, k, v, attn_metadata.prefill_length)
    #        elif self.num_kv_heads == 1:
    #            output = kernel.vanilla_prefill_mqa(q, k, v, attn_metadata.prefill_length)
    #        else:
    #            output = kernel.vanilla_prefill_gqa(q, k, v, attn_metadata.prefill_length)
    return output, kv_cache

  def _generate(
      self, q, k, v, kv_cache: KVCache, attn_metadata: AttentionMetadata
  ):
    kv_cache = self._write_generate_kv_to_kv_cache(
        k,
        v,
        kv_cache,
        attn_metadata.generate_pos,
        attn_metadata.generate_page_table,
    )

    batch = q.shape[0]

    output = kernel.decode_attention(
        q,
        kv_cache.k,
        kv_cache.v,
        attn_metadata.generate_pos,
        attn_metadata.generate_page_table,
    )

    output = output.reshape((batch, -1))

    return output, kv_cache

  def _mixed_prefill_generate(
      self, q, k, v, kv_cache: KVCache, attn_metadata: AttentionMetadata
  ):
    total_len, num_attn_heads_per_device, head_dim = q.shape
    output = jnp.zeros(
        shape=(total_len, num_attn_heads_per_device * head_dim),
        dtype=q.dtype,
    )
    padded_prompt_length = attn_metadata.prefill_pos.shape[0]
    prefill_output, kv_cache = self._prefill(
        q[:padded_prompt_length, :, :],
        k[:padded_prompt_length, :, :],
        v[:padded_prompt_length, :, :],
        kv_cache=kv_cache,
        attn_metadata=attn_metadata,
    )

    generate_output, kv_cache = self._generate(
        q[padded_prompt_length:, :, :],
        k[padded_prompt_length:, :, :],
        v[padded_prompt_length:, :, :],
        kv_cache=kv_cache,
        attn_metadata=attn_metadata,
    )

    output = jax.lax.dynamic_update_slice_in_dim(
        output,
        prefill_output,
        start_index=0,
        axis=0,
    )

    output = jax.lax.dynamic_update_slice_in_dim(
        output,
        generate_output,
        start_index=padded_prompt_length,
        axis=0,
    )

    return output, kv_cache

  def __call__(
      self, q, k, v, kv_cache: KVCache, attn_metadata: AttentionMetadata
  ):
    # q, k, v has shape as (tokens, num_heads/devices, head_dim)
    if len(attn_metadata.generate_pos.shape) == 0:
      return self._prefill(q, k, v, kv_cache, attn_metadata)
    elif len(attn_metadata.prefill_pos.shape) == 0:
      return self._generate(q, k, v, kv_cache, attn_metadata)
    else:
      return self._mixed_prefill_generate(q, k, v, kv_cache, attn_metadata)
